# -*- coding: utf-8 -*-
# (k) made-on-the-knee-of /dragon, 2017
#! /usr/bin/python3

import os, sys, stat, pwd, grp, json, uuid, re
import ctypes, platform
import shutil
import hashlib
import threading
import logging
import time

from datetime import datetime, timedelta
import importlib

try:
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass
try:
    from bsddb import db
except ImportError:
    from bsddb3 import db


def find_volume(pathname):
    """Get mount point for given path"""
    pathname = os.path.abspath(pathname)
    while not os.path.ismount(pathname):
        pathname = os.path.dirname(pathname)
    return pathname    

def get_volume_free_space(dirname):
    """
    Get free space in bytes for mount volume.
    ARG:
    dirname (string)    -- Directory or file path 
    RETURNS:
    (long)              -- Free space in bytes or None 
    """
    dirname = find_volume(dirname)
    if not dirname:
        return None
    if platform.system() == 'Windows':
        free_bytes = ctypes.c_ulonglong(0)
        ctypes.windll.kernel32.GetDiskFreeSpaceExW(ctypes.c_wchar_p(dirname), None, None, ctypes.pointer(free_bytes))
        return free_bytes.value 
    else:
        st = os.statvfs(dirname)
        return st.f_bavail * st.f_frsize 
        
def parse_date_string( str, format='datetime' ):
    """
    Datetime string parser - simple analog for dateutil.parser
    Uses hardcoded list of fornat strings.

    ARG:
    str (strng)         -- Date string
    format (enum)       -- * datetime   - return datetime object
                           * timestamp  - return float timestamp
    RETURNS:
    datetime or float timestamp or None
    """
    patterns = [ 
        "%a, %e %b %Y %H:%M:%S %z",
        "%Y-%m-%dT%H:%M:%S.%f%z",
        "%Y-%m-%dT%H:%M:%S%z",
        "%Y-%m-%dT%H:%M%z",
        "%Y-%m-%dT%H:%M:%S.%f",
        "%Y-%m-%dT%H:%M:%S",
        "%Y-%m-%dT%H:%M",
        "%Y-%m-%d",
        "%m-%e-%y %H:%M:%S",
        "%m-%e-%y %H:%M",
        "%m-%e-%y",
        "%A, %b %d",
        "%m/%d/%Y",
        "%b %e, %l:%M %p",
        "%B %Y",
        "%b %d, %Y",
        "%I:%M:%S %p",
        "%H:%M:%S",
        "%e %b %Y %H:%M:%S%p",
        "%d.%m.%y",
        "%d.%m.%Y",
        "%d-%m-%Y",
        "%d-%m-%Y %H:%M:%S",
        "%A, %d %b %Y %l:%M %p" ]
    if not str: 
        return None
    for p in patterns:
        try:
            d = datetime.strptime(str, p)
            if d:
                if 'datetime' == format:
                    return d
                elif 'timestamp' == format:
                    return (d - datetime.fromtimestamp(0)).total_seconds()
                else:
                    raise binfsValueError("DateParser: invalid format")
        except:
            pass
    return None        

def parse_interval(timestr):
    """
    Timedelta string parser. 
    ARG:
    timestr (str) in format 00.0d00.0h00.0m00.00s
    RETURNS:
    timedelta object or None
    """
    regex = re.compile(r'((?P<days>\d+?)d)'
        r'?((?P<hours>\d+?)h)'
        r'?((?P<minutes>\d+?)m)'
        r'?((?P<seconds>\d+?)s)?')
    parts = regex.match(timestr)
    if not parts:
        return None
    parts = parts.groupdict()
    time_params = {}
    for name, param in list(parts.items()):
        if param:
            time_params[name] = float(param)
    return timedelta(**time_params)
    
class binfsCommonError(Exception):
    """Common BINFS exceptions"""
    pass

class binfsValueError(ValueError):
    """Value errors for BINFS arguments"""
    pass    

class binfsFileError(IOError):
    """BINFS IO errors"""
    pass

class binfspool(list):
    """
    Pool of BINFS storages. Extends List. Implements .append .remove del iterator etc.
    Indexes storages by mount volumes.
    Implements adaptive CLEAN based on mount volume free space.
    Extended func:
        setparam    - set param for storage or default
        getparam    - get param for storage or default
        stats       - get statistics of storages
        status      - get volume status
        volumes     - get volume list
        byname      - get storage by name
        restat      - resatt all storages
        cleanall    - adaptive clean storages for each volume considering 
                      minimum free space 
    """
    DEFAULT_PARAM = {
        "minimum_space":    100000000
        ,"reduce_step_min": 0.05
        ,"reduce_step_max": 0.1
    }
    def __init__(self, name=None, logger=None):
        self.name = name
        self.param = dict(binfspool.DEFAULT_PARAM)
        self.fs = {}
        if logger:
            self.log = logger
        else:
            self.log = logging.getLogger("binfspool[%s]"%(self.name))
            self.log.addHandler(NullHandler())
        super(binfspool,self).__init__()
    def append(self,fs):
        if not isinstance(fs,binfs):
            raise binfsValueError('Only BINFS objects should be added')
        r = find_volume(fs.root)
        if not r in self.fs:
            self.fs[r] = {"param":dict(self.param), "index":[], "stats":{}, "status":"OK"}
        super(binfspool,self).append(fs)
        self.fs[r]['index'].append(self[len(self)-1])
    def remove(self,fs):
        if not isinstance(fs,binfs):
            raise binfsValueError('Only BINFS objects should be removed')
        r = find_volume(fs.root)
        if r in self.fs:
           self.fs[r]['index'].remove(fs)
        super(binfspool,self).remove(fs)   
    def __delitem__(self,i):
        fs = self[i]
        r = find_volume(fs.root)
        if r in self.fs:
           self.fs[r]['index'].remove(fs)
        super(binfspool,self).__delitem__(i) 
    def setparam(self,key,value,root=None):
        """Set parameter by name for volume by root or default"""
        if root:
            if root in self.fs:
                self.fs[root]['param'][key] = value
        else:
            self.param[key] = value
    def getparam(self,root=None):
        """Get parameter dict for volume by root or all or default"""
        return ({k:self.fs[k]['param'] for k in list(self.fs.keys())} if root=='*' 
            else (self.fs[root]['param'] if root in self.fs 
                else None)) if root else self.param
    def stats(self,root=None):
        """Get stats dict fro volume by root or all"""
        return {k:self.fs[k]['stats'] for k in list(self.fs.keys())} if not root else self.fs[root]['stats']
    def status(self):
        """Get status for all volumes"""
        return {k:self.fs[k]['status'] for k in list(self.fs.keys())}
    def volumes(self):
        """Get list of vlumes"""
        return list(self.fs.keys())
    def byname(self,name):
        """Get storage by name (first occurance)"""
        for f in self:
            if f.name == name:
                return f
        return None        
    def restat(self):
        """Force restat each storage and refresh total stats"""
        for vol,fs in self.fs.items():
            _s = {
                "freespace": get_volume_free_space(vol),
                "totalsize": 0,
                "totalcount": 0,
                "storage": []
            }
            for f in fs['index']:
                f.restat()
                _ss = f._currentstat()
                _s['totalsize'] += _ss['size']
                _s['totalcount'] += _ss['count']
                _s['storage'].append({
                    "name": f.name,
                    "root": f.root,
                    "stat": _ss
                })
            fs['stats'] = _s 
            if fs['stats']['freespace'] < fs['param']['minimum_space']:
                self.log.warning('Volume space underflow: ' + vol)
                self._clean(vol)
    def _clean(self,vol, strict='strict', destroy=False):        
        if not vol in self.fs:
            return None
        fs = self.fs[vol]
        index = fs['index']    
        smin =  fs['param']['reduce_step_min']    
        smax =  fs['param']['reduce_step_max']    
        minsz = fs['param']['minimum_space']    
        stats = fs['stats']
        bound = []
        (lbound,rbound) = (99999999,0)
        for f in index:
            minage = parse_interval(f.param['minage']).total_seconds()
            maxage = parse_interval(f.param['maxage']).total_seconds()
            if maxage > rbound: rbound = maxage
            if minage < lbound: lbound = minage
            bound.append((minage,maxage))
        n = 0
        sz = 0
        while True:
            allock = True
            for i in range(len(index)):
                f = index[i]
                (minage, maxage) = bound[i]
                age = minage + (maxage-minage)*(1 - n * (smin + (smax-smin)*(lbound - minage - lbound)/rbound))
                if age > minage:
                    allock = False
                    tr = datetime.now()-timedelta(seconds=age)
                    res = f.clean(None,tr,strict=strict,destroy=destroy)
                    f.log.info('CLEAN: %s bytes collected' % (res))
                    self.log.info("CLEAN: volume: %s storage: %s iteration: %s bytes collected: %s" % (vol, f.name, n, res))
                    sz += res
                else:
                    f.log.warning('CLEAN: age bound exceed')
                    self.log.warning("CLEAN: volume: %s storage: %s iteration: %s AGE BOUND EXCEED" % (vol, f.name, n))
            if minsz < get_volume_free_space(vol):
                fs['status'] = 'OK'
                break
            else:
                fs['status'] = 'LOW'
            if allock:
                self.log.error('CLEAN: volume: %s CRITICAL SPACE UNDERFLOW!' % (vol))
                fs['status'] = 'CRITICAL'
                break
            n += 1    
        return sz
    def cleanall(self, strict='strict', destroy=False):
        """Force clean all storages considering its min-max ages and required free space fro volume"""
        sz = 0
        for v in self.volumes():
            sz += self._clean(v,strict=strict,destroy=destroy)
        return sz    

class binfs:
    """
    Represents a single BINFS storage identified by root directory.
    Optimized for binary file storage with metadata indexed in BerkeleyDB
    Exports functionalities:
        - storage manipulation
        - statistics
        - clean and maintenance
    """
    DEFAULT_PARAM = {
         "ts_to_utc": True
        ,"meta_page_size": 512
        ,"class": "timestamp"
        ,"opt": ["%Y","%m","%d","%H"]
        ,"minage": '1h'
        ,"maxage": '7d'
    }
    LOCK = {}
        
    def __init__(self, name = "", root = None, logger = None, oCreate = False, oTruncate = False, oRepair=True, oParam = None, setUID = None, setGID = None ):
        """
        BINFS object constructor

        ARG:
            root    (string)            -- Root directory (not null, for one storage ONLY)
            name    (string)            -- Instane name for logging ident
            logger  (logging.Logger)    -- Logger (optional), default logger in Null
            setUID      (int or str)    -- Use SETUID (root only)
            setGID      (int or str)    -- Use SETGID (root only)
            oCreate     (bool)          -- Create storage root if not exists
            oTruncate   (bool)          -- Clear and drop storage if exists 
            oParam      (dict)          -- Parameters extends the default
                Parameters:
                -----------
                    "class"             -- storage key class = ['timestamp','integer','string']
                    "opt"               -- list of subdir format options special for class
                    "meta_page_size"    -- page size for BDB metadata storage
                    "ts_to_utc"         -- convert key timestamp to UTC for timestamp class 

        """
        if not root:
            raise binfsValueError('Root dir not set')
        self.root = root
        self.name = name
        self.param = dict(binfs.DEFAULT_PARAM)
        self.paramchk = {}
        self.stat = {}
        if not self.root in binfs.LOCK:
            binfs.LOCK[self.root] = (threading.Lock(), threading.Condition()) 
        if logger:
            self.setlogger(logger)
        else:
            self.log = logging.getLogger("-.binfs[%s]"%(self.name))
            self.log.addHandler(NullHandler())
        self.owner = ( 
            -1 if not setUID else ( setUID if isinstance(setUID,int) else pwd.getpwnam(setUID).pw_uid ),
            -1 if not setGID else ( setGID if isinstance(setGID,int) else grp.getgrnam(setGID).gr_gid )
        )    
        if oParam:
            for k,v in oParam.items():
                self.param[k] = v
        if os.path.exists(self.root):
            while True:        
                if oTruncate:
                    def on_ro(func, dirpath, _):   
                        os.chmod(dirpath,stat.S_IWRITE)
                        func(dirpath)
                    if os.path.isdir(self.root):
                        shutil.rmtree(self.root, onerror=on_ro)
                    else:
                        os.remove(self.root, onerror=on_ro)
                    oldumask = os.umask(0)
                    os.makedirs(self.root)
                    os.chown(self.root,self.owner[0],self.owner[1])
                    os.umask(oldumask)
                    self._init()
                    self._saveconf(force=True)   
                    break
                else:
                    if not os.path.isdir(self.root):
                        raise binfsFileError('Root is not a dir')
                    if not self._loadconf(repair=oRepair):
                        if oCreate:
                            oTruncate = True
                            continue
                        else:
                            raise binfsFileError("Error in FS structure cannot be repaired")
                    break 
        else:
            if oCreate:
                oldumask = os.umask(0)
                os.makedirs(self.root)
                os.chown(self.root,self.owner[0],self.owner[1])
                os.umask(oldumask)
                self._init()
                self._saveconf(force=True)    
            else:
                raise binfsCommonError('Root dir doesn`t exists')

    def setlogger(self,logger):
        """Sets the bings.log to given logging.Logger"""
        self.log = logger        
        
    def _init(self):
        self.param['uuid'] = str(uuid.uuid4())
        self.param['ctime'] = self.param['atime'] = datetime.utcnow().isoformat()
    
    def _locklock( self, timeout=None ):
        if not self.root in binfs.LOCK:
            raise binfsCommonError("Lock: invalid lock object: " + self.root)
        (lock,cond) = binfs.LOCK[self.root]
        with cond:
            if timeout:
                ctime = stime = time.time()
                while ctime < stime + timeout:
                    if lock.acquire(False):
                        return True
                    else:
                        cond.wait(timeout - ctime +stime)
                        ctime = time.time()                
            else:
                lock.acquire(True)
                return True
        return False

    def _lockrelease(self):
        if not self.root in binfs.LOCK:
            raise binfsCommonError("Lock: invalid lock object: " + self.root)
        (lock,cond) = binfs.LOCK[self.root]
        if lock.locked():
            with cond:
                lock.release()
                cond.notify()
            
    def _paramchk(self):
        buf = json.dumps(self.param)
        sha = hashlib.sha256(buf).hexdigest()
        size = len(buf)
        res = ( 'SIZE' in self.paramchk 
            and self.paramchk['SIZE'] == size 
            and 'SHA256' in self.paramchk 
            and self.paramchk['SHA256'] == sha )
        self.paramchk['SHA256'] = sha
        self.paramchk['SIZE'] = size 
        self.paramchk['GUID'] = self.param['uuid']
        self.paramchk['UP'] = datetime.utcnow().isoformat()
        return res    
    
    def getparam(self,key):
        """Get BINFS parameter by key""" 
        return self.param[key] if key in self.param else None
           
    def setparam(self,key,val):
        """Set BINFS paraemter by key. Create if not exists"""
        self.param[key] = val
        self._saveconf()

    def _saveconf(self, force = False):
        if not self._paramchk() or force:
            eq = False
            if not force:
                try:
                    with open(os.path.join(self.root,'__binfs.check'),'r') as _fcheck:
                        check = json.load(_fcheck)
                        eq = ( 'SIZE' in self.paramchk and 'SIZE' in check
                            and self.paramchk['SIZE'] == check['SIZE']
                            and 'SHA256' in self.paramchk and 'SHA256' in check
                            and self.paramchk['SHA256'] == check['SHA256'] )
                except IOError:
                    force = True
            if not eq or force:    
                with open(os.path.join(self.root,'__binfs'),'w') as _fconf, open(os.path.join(self.root,'__binfs.check'),'w') as _fcheck:
                    json.dump(self.param,_fconf,sort_keys=True, indent=4, ensure_ascii=False)
                    json.dump(self.paramchk,_fcheck,sort_keys=True, indent=4, ensure_ascii=False)
                os.chown(os.path.join(self.root,'__binfs'),self.owner[0],self.owner[1])
                os.chown(os.path.join(self.root,'__binfs.check'),self.owner[0],self.owner[1])
    def _loadconf(self, repair=True):
        try:        
            with open(os.path.join(self.root,'__binfs'),'r') as _fconf:
                self.param = json.load(_fconf)
        except Exception as ex:
            self.log.error('Invalid config: ' + str(ex))
            return False
        try:   
            with open(os.path.join(self.root,'__binfs.check'),'r') as _fcheck:
                self.paramchk = json.load(_fcheck)
        except IOError:
            if repair:
                self.restat()
                self._saveconf(force=True)
                return True
            else:
                return False
        except Exception as ex:
            self.log.error('Config check error: ' + str(ex))
            return False
        if not self._paramchk():
            self._saveconf(force=True)
        self.stat = self._currentstat()
        return True

    """
        STAT FUNCTIONALITY
    """        
        
    def _currentstat(self):
        try:
            with open(os.path.join(self.root,'__binfs.stat'),'r') as _fstat:
                _s = json.load(_fstat)
                return _s[max(_s.keys())]
        except IOError:
            return {}
        except ValueError as ex:
            self.log.error("CurrentStat: read error: "+ str(ex))
            self._savestat()
            
    def _savestat(self):
        _s = {datetime.utcnow().isoformat():self.stat}
        try:
            with open(os.path.join(self.root,'__binfs.stat'),'r') as _fstat:
                _s.update(json.load(_fstat))
        except IOError:
            pass
        except ValueError as ex:
            self.log.error("Stat: read error: " + str(ex))
            pass
        with open(os.path.join(self.root,'__binfs.stat'),'w') as _fstat:
            json.dump(_s, _fstat, sort_keys=True, indent=4, ensure_ascii=False)
        os.chown(os.path.join(self.root,'__binfs.stat'),self.owner[0],self.owner[1])

    def _testdir(self, dirpath):
        if not os.path.exists(dirpath):
            try:
                os.makedirs(dirpath)
            except Exception as ex:
                self.log.error("DIRTest: %s : %s " % (dirpath, str(ex)))
                return False
        else:
            try:
                td = os.path.join(dirpath, '__binfs.' + str(os.getpid()))
                if os.path.exists(td):
                    os.rmdir(td)
                os.makedirs(td)
            except Exception as ex:
                self.log.error("DIRTest: %s : %s " % (dirpath, str(ex)))
                return False
        return True
        
    def _statdir(self, ipath, idepth = 0):    
        def _statappend(ito, ifrom):
            ito['ocount'] += ifrom['ocount']
            ito['osize'] += ifrom['osize']
            if not ito['oftimemin'] or ifrom['oftimemin'] < ito['oftimemin']:
                ito['oftimemin'] = ifrom['oftimemin']
            if not ito['oftimemax'] or ifrom['oftimemax'] > ito['oftimemax']:
                ito['oftimemax'] = ifrom['oftimemax']
            if not ito['opathmin'] or ifrom['opathmin'] < ito['opathmin']:
                ito['opathmin'] = ifrom['opathmin']
            if not ito['opathmax'] or ifrom['opathmax'] > ito['opathmax']:
                ito['opathmax'] = ifrom['opathmax']
        def _dirproc(ipath, idepth):
            stsum = {
                "ocount": 0,
                "osize": 0,
                "oftimemin": None,
                "oftimemax": None,
                "opathmin": None,
                "opathmax": None
            }
            stcurrent = {
                "ocount": 0,
                "osize": 0,
                "oftimemin": None,
                "oftimemax": None,
                "opathmin": None,
                "opathmax": None
            }
            if not os.path.exists(ipath):
                return stsum
            rescan = True
            if idepth > 0:
                for f in os.listdir(ipath):
                    if f.startswith('.'):
                        continue
                    if f.startswith('__binfs'):
                        continue
                    if os.path.isdir(os.path.join(ipath, f)):
                        stats = _dirproc(os.path.join(ipath, f), idepth - 1)
                        _statappend(stsum, stats)
            else:
                pathdir, pathname = os.path.split(ipath)
                st = os.lstat(ipath)
                sto = dict((name, getattr(st, name)) for name in dir(st) if
                           not name.startswith('__'))
                sth = "%x" % (hash(tuple(sorted(hash(x) for x in list(sto.items())))) &
                              0xffffffffffffffff)
                stfile = os.path.join(pathdir, '.stat.{0}.{1}'.format(pathname, sth))
                if os.path.exists(stfile):
                    try:
                        with open(stfile, 'r') as stfd:
                            stf = json.load(stfd)
                        rescan = False
                        stcurrent = stf['stats']
                    except Exception as ex:
                        self.log.error("stat prescan " + str(ex))
                        rescan = True
                        os.remove(stfile)
                if rescan:
                    try:
                        for f in os.listdir(pathdir):
                            if f.startswith('.stat.' + pathname):
                                os.unlink(os.path.join(pathdir, f))
                        for f in os.listdir(ipath):
                            if f.startswith('.'):
                                continue
                            if f.startswith('__binfs'):
                                continue
                            sst = os.lstat(os.path.join(ipath, f))
                            cstats = {
                                "ocount": 1,
                                "osize": sst.st_size,
                                "oftimemin": sst.st_ctime,
                                "oftimemax": sst.st_ctime,
                                "opathmin": f,
                                "opathmax": f
                            }
                            _statappend(stcurrent, cstats)
                        with open(stfile, 'w') as stfd:
                            json.dump({"stat": sto, "stats": stcurrent}, stfd,
                                      sort_keys=True, indent=4, ensure_ascii=False)
                        os.chown(stfile,self.owner[0],self.owner[1])              
                    except Exception as ex:
                        self.log.error(str(ex))
                        raise
                _statappend(stsum, stcurrent)
            return stsum
        return _dirproc(ipath,idepth)

    def getstat(self):
        """Returns current status dict"""
        return self._currentstat()

    def restat(self):
        """
        Force recalculate storage stats. May affect disk overload by storage dierctory loop.
        Optimized to skip untouched directories on low tree level.
        """
        if self._locklock(timeout=2):
            stats = self._statdir(self.root, len(self.param['opt']))
            self._lockrelease()
            _s =  {
                "uuid":            self.param['uuid'],
                "moment":          datetime.utcnow().isoformat(),
                "freespace":       get_volume_free_space(self.root),
                "size":            stats['osize'],
                "count":           stats['ocount'],
                "ftimemin":        None if not 'oftimemin' in stats or not stats['oftimemin'] else (
                    datetime.utcfromtimestamp(float(stats['oftimemin'])).isoformat()
                    if self.param['ts_to_utc'] else ( 
                        (datetime.fromtimestamp(float(stats['oftimemin']))).isoformat() if stats['oftimemin'] else None )),
                "ftimemax":        None if not 'oftimemax' in stats or not stats['oftimemax'] else (
                    datetime.utcfromtimestamp(float(stats['oftimemax'])).isoformat() 
                    if self.param['ts_to_utc'] else (
                        (datetime.fromtimestamp(float(stats['oftimemax']))).isoformat() if stats['oftimemax'] else None )),
                "pathmin":         stats['opathmin'],
                "pathmax":         stats['opathmax']
            }    
            self.stat = _s
            self._savestat()
        else:
            self.log.error("Restat: duplcate locked")

    """
        CLEAN FUNCTIONALITY
    """            

    def _rmtree(self, filepath):
        if not os.path.exists(filepath):
            return False
        if os.path.isdir(filepath):
            shutil.rmtree(filepath)
        else:
            os.remove(filepath)
        return True
    
    def _destroy(self, confirm=False):
        if confirm:
            shutil.rmtree(self.root)
            self.root = None
            self.param = self.paramchk = self.stat = {}
            self.log.info('Storage is destroyed')

    def clean(self, ifrom, ito, mode='default', strict='strict', destroy=False ):
        """
        Force to delete storage data of key range.

        ARG:
        mode        (enum)  -- key bound mode
                               * integer
                               * timestamp
                               * filetime
                               * regexp
                               * (default) - as key class for time or int
        ifrom       (any)   -- left range bound (timestamp, long, float, string)
        ito         (any)   -- right  range bound (None = open range) 
        strict      (enum)  -- how to round edges:
                                * strict -- delete strictly between given keys (default)
                                * around -- round edge to the low level directory
        """
        ret = -1
        param = {}
        mdb = db.DB()
        mdb.close()
        metapath = ""
        root = self.root
        kind = self.param['class']
        opt = self.param['opt']
        mdep = len(opt)
        ispace = get_volume_free_space(self.root)
        if not mode or mode == 'default':
            if kind in ['timestamp','integer']:
                mode = kind
            else:
                raise binfsValueError('Clean: Invalid key mode')
        if 'filetime' == mode:
            """  don`t forget TODO it !!!! """
            pass
        elif 'regexp' == mode:
            """  don`t forget TODO it !!!! """
            pass
        elif 'timestamp' == mode or 'integer' == mode:
            ifp = itp = None
            if ifrom:
                ifp = self.path(ifrom)
                if not ifp:
                    raise binfsValueError('Clean: invalid FROM')
            path_from = ifp
            if ito:
                itp = self.path(ito)
                if not itp:
                    raise binfsValueError('Clean: invalid TO')
            path_to = itp 
            if not ifp and not itp:
                try:
                    self._destroy(confirm=destroy)
                except OSError as oe:
                    self.log.error("Clean: " + str(oe))
                    pass
            else:
                depth = -1
                while True:
                    if (path_from and not path_to and path_from <= root) or (
                            path_to and not path_from and path_to <= root) or (
                            path_from and path_to and path_from == path_to):
                        break
                    if 'around' == strict and depth < 0:
                        depth += 1
                        if path_from:
                            path_from = os.path.dirname(path_from)
                            try:
                                shutil.rmtree(path_from)
                            except OSError as oe:
                                self.log.error("Clean: " + str(oe))
                                pass
                            dir_from,name_from = os.path.split(path_from)
                            if depth < 1:
                                metapath = os.path.join(dir_from,".%s..meta"%(name_from))
                                try:
                                    os.unlink(metapath)
                                except OSError as oe:
                                    self.log.error("Clean: " + str(oe))
                                    pass
                        if path_to:
                            path_to = os.path.dirname(path_to)
                            try:
                                shutil.rmtree(path_to)
                            except OSError as oe:
                                self.log.error("Clean: " + str(oe))
                                pass
                            dir_to,name_to = os.path.split(path_to)
                            if depth < 1:
                                metapath = os.path.join(dir_to,".%s..meta"%(name_to))
                                try:
                                    os.unlink(metapath)
                                except OSError as oe:
                                    self.log.error("Clean: " + str(oe))
                                    pass
                    if path_from:
                        dir_from, name_from = os.path.split(path_from)
                    if path_to:
                        dir_to, name_to = os.path.split(path_to)
                    if path_from:
                        if os.path.exists(dir_from):
                            has_meta = False
                            if depth < 0:
                                try:
                                    mdir, mblock = os.path.split(dir_from)
                                    mdb.close()
                                    mdb = db.DB()
                                    mdb.open(os.path.join(mdir,".%s..meta"%(mblock)), dbtype=db.DB_BTREE )
                                    os.chown(os.path.join(mdir,".%s..meta"%(mblock)),self.owner[0],self.owner[1])
                                    has_meta = True
                                except db.DBNoSuchFileError:
                                    has_meta = False
                                    pass
                            try:
                                for f in os.listdir(dir_from):
                                    if f > name_from and not (path_to and
                                                              dir_to == dir_from and
                                                              f >= name_to):
                                        pth = os.path.join(dir_from, f)
                                        try:
                                            if os.path.isdir(pth):
                                                shutil.rmtree(pth)
                                                if depth < 1:
                                                    metapath = os.path.join(dir_from,".%s..meta"%(f))
                                                    os.unlink(metapath)
                                            else:
                                                if depth < 0:
                                                    os.unlink(pth)
                                                    if has_meta: 
                                                        mdb.delete(f)
                                        except OSError as oe:
                                            self.log.error("Clean: " + str(oe))
                                            continue
                                        except db.DBNotFoundError: 
                                            continue    
                            except OSError as oe:
                                self.log.error("Clean: " + str(oe))
                                pass
                    if path_to:
                        if os.path.exists(dir_to):
                            has_meta = False
                            if depth < 0:
                                try:
                                    mdir, mblock = os.path.split(dir_to)
                                    mdb.close()
                                    mdb = db.DB()
                                    mdb.open(os.path.join(mdir,".%s..meta"%(mblock)), dbtype=db.DB_BTREE )
                                    os.chown(os.path.join(mdir,".%s..meta"%(mblock)),self.owner[0],self.owner[1])
                                    has_meta = True
                                except db.DBNoSuchFileError:
                                    has_meta = False
                                    pass
                            try:
                                for f in os.listdir(dir_to):
                                    f = f.encode('utf-8')
                                    if f < name_to and not (path_from and
                                                            dir_from == dir_to and
                                                            f <= name_from):
                                        pth = os.path.join(dir_to, f)
                                        try:
                                            if os.path.isdir(pth):
                                                shutil.rmtree(pth)
                                                if depth < 1:
                                                    metapath = os.path.join(dir_to,".%s..meta"%(f))
                                                    os.unlink(metapath)
                                            else:
                                                if depth < 0:
                                                    os.unlink(pth)
                                                    if has_meta: 
                                                        mdb.delete(f)
                                        except OSError as oe:
                                            self.log.error("Clean: " + str(oe))
                                            continue
                                        except db.DBNotFoundError: 
                                            continue    
                            except OSError as oe:
                                self.log.error("Clean: " + str(oe))
                                pass
                    if path_from:
                        path_from = dir_from
                    if path_to:
                        path_to = dir_to
                    depth += 1
        else:
            self.log.error("Clean: invalid clean mode")
        return get_volume_free_space(self.root) - ispace
            
    def purge(self):
        """
        Garbadge collecor.
        Force to delete empty directories. 
        """
        def _dirproc(ipath, idepth):
            res = []
            if not os.path.exists(ipath):
                return res
            if idepth < 0:
                return res
            if idepth > 0:
                cnt = 0
                for f in os.listdir(ipath):
                    if f.startswith('.'):
                        continue
                    if f.startswith('__binfs'):
                        continue
                    if os.path.isdir(os.path.join(ipath, f)):
                        res += _dirproc(os.path.join(ipath, f), idepth - 1)
                        if os.path.exists(os.path.join(ipath, f)):
                            cnt += 1
                if cnt == 0:
                    shutil.rmtree(ipath, ignore_errors=True)
                    res.append(ipath)
            else:
                st = os.lstat(ipath)
                sto = dict((name, getattr(st, name)) for name in dir(st) if
                           not name.startswith('__'))
                sth = "%x" % (hash(tuple(sorted(hash(x) for x in list(sto.items())))) &
                              0xffffffffffffffff)
                rescan = True
                pure = False
                idir, iname = os.path.split(ipath)
                stfile = os.path.join(idir, '.stat.' + iname + '.' + str(sth))
                if os.path.exists(stfile):
                    try:
                        with open(stfile) as stfd:
                            stf = json.load(stfd)
                        if stf['stats']['ocount'] == 0:
                            pure = True
                            rescan = False
                    except Exception as ex:
                        self.log.error("Purge: " + str(ex))
                if rescan:
                    if not os.listdir(ipath):
                        pure = True
                try:
                    if pure:
                        shutil.rmtree(ipath, ignore_errors=True)
                        res.append(ipath)
                except Exception as ex:
                    self.log.error("Purge: " + str(ex))
            return res
        return tuple(_dirproc(self.root, len(self.param['opt'])))        
                    
    """
        STORAGE FUNCTIONALITY
    """            

    def load( self, ikey ):
        """Returns the binary data of given key or None"""
        _path = self.path(ikey)
        if _path:
            with open(_path,'rb') as file:
                return file.read()
        return None        
    
    def save(self, ikey, idata, overwrite=True):
        """
        Stores the binary data of given key.

        ARG:
        ikey        (any)   -- storage key
        idata       (any)   -- data to store (to be a binary string)
        overwrite   (bool)  -- delete existing data or raise data error (default: True)

        RETURNS:
        int                 -- number of bytes stored or None
        """
        _path = self.path(ikey)
        try:
            self._prepare(_path, mkdir=True, remove=overwrite)
        finally:
            with open(_path,'wb') as file:
                file.write(idata)
                return len(idata)
    
    def drop(self, ikey):
        """Force to delete binary data for given key"""
        return self._rm(ikey)

    def meta_get(self, ikey):
        """Returns metdata for given key or None""" 
        return self._meta_get(self.path(ikey))
        
    def _meta_get(self, ipath):
        if not ipath: return None
        _r = None
        _t = ipath.encode('utf-8')
        [_t,key] = os.path.split(_t)
        [dir,block] = os.path.split(_t)
        _db = db.DB()
        try:
            _db.open(os.path.join(dir,".%s..meta"%(block)), dbtype=db.DB_BTREE, flags=db.DB_RDONLY)
            os.chown(os.path.join(dir,".%s..meta"%(block)),self.owner[0],self.owner[1])
            _r = _db.get( key )
        except db.DBNotFoundError:
            return None
        except Exception as ex:    
            self.log.error("MetaGet: " + str(ex))
            return None
        return _r 
    
    def meta_set(self, ikey, idata, overwrite=True):
        """
        Stores metadata for given key.
        ARG:
        ikey        (any)   -- storage key
        idata       (any)   -- data to store (all types to be serializeable to JSON)
        overwrite   (bool)  -- delete existing data or raise data error (default: True)
        """
        return self._meta_set(self.path(ikey), idata, overwrite)
        
    def _meta_set(self, ipath, idata, overwrite=True):
        if not ipath: return None    
        _r = None
        _t = ipath
        [_t,key] = os.path.split(_t.encode('utf-8'))
        [dir,block] = os.path.split(_t)
        _db = db.DB()
        _f = os.path.join(dir,".%s..meta"%(block))
        _db.set_pagesize(self.param['meta_page_size'])
        try:
            _db.open(os.path.join(dir,".%s..meta"%(block)), dbtype=db.DB_BTREE, flags=db.DB_CREATE)
            os.chown(os.path.join(dir,".%s..meta"%(block)),self.owner[0],self.owner[1])
            _db.put(key, idata, flags = db.DB_NOOVERWRITE if not overwrite else 0 )
            _r = len(idata)
        except Exception as ex:    
            self.log.error("MetaSet: " + str(ex))
            raise
        return _r

    def meta_drop(self, ikey):
        """Delete metadata for given key"""
        return self._meta_drop(self.path(ikey))
        
    def _meta_drop(self, ipath):
        if not ipath: return None
        _r = None
        _t = ipath.encode('utf-8')
        [_t,key] = os.path.split(_t)
        [dir,block] = os.path.split(_t)
        _db = db.DB()
        try:
            _db.open(os.path.join(dir,".%s..meta"%(block)), dbtype=db.DB_BTREE )
            os.chown(os.path.join(dir,".%s..meta"%(block)),self.owner[0],self.owner[1])
            _db.delete( key )
            _r = 1
        except db.DBNotFoundError:
            _r = 0
        except Exception as ex:    
            self.log.error("MetaDrop: " + str(ex))
            return None
        return _r 

    def _prepare(self, filepath, mkdir=True, remove=True):
        if os.path.exists(filepath):
            if remove:
                os.remove(filepath)
        else:        
            folder = os.path.dirname(filepath)
            if not os.path.isdir(folder):
                if os.path.exists(folder):
                    os.remove(folder)
                if mkdir:
                    oldumask = os.umask(0)
                    os.makedirs(folder)
                    os.chown(folder,self.owner[0],self.owner[1])
                    os.umask(oldumask)
                else:
                    raise binfsFileError("Dir doesn`t exists: " + folder)
        
    def _exists(self, filepath):
        return os.path.exists(filepath)
    
    def _rm(self, filepath):
        try:
            os.remove(filepath)
        except:
            return False
        return True    

    def path(self, ikey):
        """
        Determine the binary file path for the given key.

        ARG:
        key (any)           -- storage key of the given class

        RETURNS:
        path (string)       -- binary file absolute path

            Opt format for classes:
            -----------
            timestamp:
                array of strftime formats ['%Y','%M','....'], file is of ts float
            string:
                ['separator',['the{0}','a{1}']] - split separator and list of format strings
            bytes:
                [1,-1,2] - list of byte numbers starting with 1 (-1) - from beginning or end of hex string
                hex formed from decimal,hex or uuid
            integer:
                [10,[4,3,2,]] - low mask digits in dec
                [16,[4,3,2,]] - low mask bytes in hex
        """ 
        path = None
        root = self.root
        opt = self.param['opt']
        if   'timestamp' == self.param['class']:
            dt = None
            if isinstance(ikey,datetime):
                tf = (ikey - datetime.fromtimestamp(0)).total_seconds()
                dt = datetime.utcfromtimestamp(tf) if self.param["ts_to_utc"] else datetime.fromtimestamp(tf)
            if isinstance(ikey,float) or (isinstance(ikey,str) and ikey.find('.') > 0):
                try:
                    tf = float(ikey)
                    dt = datetime.utcfromtimestamp(tf) if self.param["ts_to_utc"] else datetime.fromtimestamp(tf)
                except ValueError:    
                    dt = None
            if dt is None:
                try:
                    tf = int(ikey) / 1e6
                    dt = datetime.utcfromtimestamp(tf) if self.param["ts_to_utc"] else datetime.fromtimestamp(tf)
                except ValueError:
                    dt = None
            if dt is None:
                try:
                    tf = parse_date_string(ikey, format='timestamp') 
                    dt = datetime.utcfromtimestamp(tf) if self.param["ts_to_utc"] else datetime.fromtimestamp(tf)
                except Exception as ex:
                    self.log.error("Path: timestamp : " + str(ex))
                    return None
            path = root
            for part in opt:
                path = os.path.join(path,dt.strftime(part))
            path = os.path.join(path,"{:.6f}".format(tf))
            
        elif 'string' == self.param['class']:
            delim = opt[0]
            parts = [x.strip() for x in ikey.split(delim)]
            path = root
            for part in opt[1]:
                path = os.path.join(path, part.format(parts))
            path = os.path.join(path, ikey)

        elif 'bytes' == self.param['class']:
            val = None
            try:
                val = "%X" % int(round(float(ikey)))
                if len(val) & 1 == 1:
                    val = '0' + val
            except ValueError:
                val = None
            if not val:
                try:
                    val = "%X" % int(ikey, 16)
                    if len(val) & 1 == 1:
                        val = '0' + val
                except ValueError:
                    val = None
            if not val:
                try:
                    val = uuid.UUID(ikey).hex.upper()
                    if len(val) & 1 == 1:
                        val = '0' + val
                except ValueError:
                    val = None
            if not val:
                val = ikey
            path = root
            for i in opt:
                u = i * 2 - 1
                if u == 0:
                    u = None
                l = i * 2 + 1
                if l == 0:
                    l = None
                path = os.path.join(path, val[l:u])
            path = os.path.join(path, ikey)

        elif 'integer' == self.param['class']:
            val = None
            try:
                val = int(val, 10)
            except ValueError:
                val = None
            if not val:
                try:
                    val = int(ikey, 16)
                except ValueError:
                    val = None
            if not val:
                self.log.error("Path: integer: invalid key")
                return None
            else:
                if opt[0] == 10:
                    path = root
                    for d in opt[1]:
                        x = val - val % pow(10, d)
                        path = os.path.join(path, "{0:020d}".format(x))
                    path = os.path.join(path, "{0:020d}".format(val))
                elif opt[0] == 16:
                    path = root
                    for d in opt[1]:
                        x = val >> d * 8 << d * 8
                        path = os.path.join(path, "{0:016X}".format(x))
                    path = os.path.join(path, "{0:016X}".format(val))
                else:
                    self.log.error("Path: integer: invalid opt")
                    return None
        else:
            self.log.error("Path: invalid key class")
            return None
        return path

        
