create sequence if not exists eventstore_seq;

do language plpgsql $$
begin
if current_setting('server_version_num')::integer < 90600 then
    create or replace function to_regclass(text) returns regclass language sql as 'select to_regclass($1::cstring)';
end if;
end;
$$;


--
--  F: pre-moderator, validator
--

create or replace function public._event_propose(idata jsonb, out errcode integer, out errmsg varchar, out result jsonb) as $$
declare
    r record;
    plate text;
    duptimeout float;
    cnt integer;
begin
    errcode := 0;
    errmsg := '';
    
    -- Duplicate cutter
    plate :=  (idata #>> '{measurements,0,waypoints,best,vehicle,license-plates,0,text,ucode}');
    for r in select pattern,value,elen,estimate_patternclause(pattern) as clause from estimate_details('duptimeout',idata::json) order by elen desc loop
        duptimeout := r.value::text::float;
        execute format('select count(1) from eventindex where %s and plate = $1 and ctime > now() - interval ''1 sec'' * $2',r.clause) using plate, duptimeout into cnt;
        if cnt > 0 then
            errcode := 2;
            errmsg := 'DUPLICATE: '||r.pattern::text;
            return;
        end if;
    end loop;
    
    
    -- TBD premoderation filters
end;
$$ language plpgsql;

--
--  T: preparator, transformator
--

create or replace function public._event_prepare( idata jsonb ) returns public.eventstore as $$
declare
    new public.eventstore;
    _errcode integer;
    _errmsg varchar;
    _res jsonb;
begin
    new.id = nextval('eventstore_seq');
    new.event := idata -> 'event';
    new.dist := idata -> 'dist';
    new.store := idata -> 'dist' ->> 'store';
    if new.store isnull or new.store = '' then
        new.store := '_default';
    end if;    
    new.flag := (idata ->> 'flag')::integer;
    select into _errcode, _errmsg, _res * from public._event_propose(new.event);
    if not _errcode = 0 then
        raise exception 'E:PROP: [%] %', _errcode, _errmsg using hint='400';
    end if;
    if _o('event_transform_on_insert')::boolean then
        new.event := _event_transform(new.event);
    end if;    
    new.guid := (new.event ->> 'uuid')::uuid;
    if new.guid is null then
        raise exception 'E:P: Invalid guid: %',new.event using hint='400';
    end if;    
    new.ctime := (new.event ->> 'time')::timestamptz;
    if new.ctime is null then
        raise exception 'E:P: Invalid time stamp' using hint='400';
    end if;
    new.origin := (new.event -> 'origin' ->> 'uuid');
    new.doctype := (new.event ->> 'type');
    if new.doctype is null then
        raise exception 'E:P: Invalid type' using hint='400';
    end if;
    new.scrollid := (new.event ->> 'scrollid')::bigint;
    if new.scrollid is null then
        new.scrollid := (extract(epoch from date_trunc('second',new.ctime))*1000000)::bigint + mod(NEW.id,100000);
    end if;    
    return new;
end;
$$ language plpgsql;


create or replace function public._event_transform(idata jsonb) returns jsonb as $$
import json, re

return idata
$$ language plpython3u;


--
--  T: postfactum trigger
--

create or replace function public._event_postfactum( new public.eventstore, out _errcode integer, out _errmsg varchar) as $$
declare
    r record;
    st timestamptz;
begin
    _errcode := 0;
    _errmsg := 'OK';
    for r in 
        select * 
        from public.postfactum 
        where 
            enabled and 
            section = 'event' 
    loop
        begin
            st := clock_timestamp();
            execute 
                'select * from '||r.handler||'($1)' 
            into 
                _errcode, _errmsg
            using
                new;
            if _errcode <> 0 then 
                perform core.log('postfactum: '||r.handler,'error',format('[%s] %s',_errcode,_errmsg)); 
                return;
            end if;    
        exception when others then
            _errcode := -10;
            _errmsg := E'FATAL: {state:'||SQLSTATE||E',errm:'||SQLERRM||E'}';
            perform core.log('postfactum: '||r.handler,'error',_errmsg);
            return;
        end;    
    end loop;    
end; 
$$ language plpgsql;


--
--  F:P: attribute collector
--

create or replace function public.postfactum_event_collect(new public.eventstore, out _errcode integer, out _errmsg varchar) as $$
declare
    _i integer;
    _j jsonb;
    _s text;
begin
    _errcode := 0; _errmsg := 'OK';

    -- collect origin
    _j = new.event -> 'origin';
    if _j is null or _j ->> 'uuid' is null then
        _errcode := 1; _errmsg = 'C:Origin: Invalid origin';
        return;
    end if;    
    insert into public.collection_event_origin  
        (id, name, origin) 
    values ( 
        (_j ->> 'uuid')::uuid,
        _j ->> 'name',
        _j::jsonb
    ) on conflict (id)
    do update 
        set origin = _j::jsonb, 
            name = _j ->> 'name'
        where public.collection_event_origin.id = (_j ->> 'uuid') :: uuid 
    ;

    -- collect lane
    _s = new.event #>> '{measurements,0,waypoints,best,location,lane}'; 
    if _s notnull then
        insert into public.collection_event_lane
            (origin, lane)
        select     
            (_j ->> 'uuid')::uuid as origin,
            _s as lane
        on conflict (origin,lane)
        do nothing
        ;
    end if;
    
    -- collect profile
    _j = new.event #> '{evidence,args,profile}';
    if _j notnull then
        insert into public.collection_event_profile 
            (doctype, id,  name, profile) 
        values (
                new.doctype,
                _j ->> 'id',
                _j ->> 'name',
                _j
        ) on conflict (doctype,id)
        do update
            set name = _j ->> 'name', profile = _j 
            where public.collection_event_profile.doctype = new.doctype and  public.collection_event_profile.id = _j ->> 'id' 
        ;    
    end if;
end;
$$ language plpgsql;


create or replace function public._event_reindex( new public.eventstore, clear boolean default false, flag integer = 0 ) returns void as $$
declare    
    _doclass varchar(32);
    _sc text;
    _sv text;
    _sp integer;
    _dir text;
    _t text;
    _j jsonb;
    _g text[];
    r record;
begin
    if clear then
        delete from public.eventindex where id = new.id;
    end if;    
    _doclass := estimate('doclass',json_build_object('type',new.doctype));
    _sp := null;
    if _sp isnull then
        _sp := (new.event #>> '{evidence,measurements,section-speed}')::real::integer;
    end if;
    if _sp isnull then
        _sp := (new.event #>> '{evidence,measurements,speed,radar}')::real::integer;
    end if;
    if _sp isnull then
        _sp := (new.event #>> '{measurements,0,waypoints,best,vehicle,section-speed,measured}')::real::integer;
    end if;    
    if _sp isnull then
        _sp := (new.event #>> '{measurements,0,waypoints,best,vehicle,spot-speed,radar,measured}')::real::integer;
    end if;
    _dir := (new.event #>> '{measurements,0,waypoints,best,vehicle,dir}');
    if _dir isnull then
        _dir := 'unknown';
    end if;   
    select into _g array(select value ->> 'name' from jsonb_array_elements(new.event #> '{alarms}'));
    execute format (
        'insert into public.eventindex (   
            id,
            flag,
            store,
            scrollid,
            guid,
            pguid,
            ctime,
            wtime,
            stime,
            etime,
            origin,
            lane,
            doctype,
            doclass,
            docprofile,
            plate,
            country,
            region,
            speed,
            dir,
            make,
            model,
            mm,
            class,
            
            alarms
         ) values (
            $1,
            $2,
            $3,
            $4,
            $5,
            $6,
            $7,
            $8,
            $9,
            $10,
            $11,
            $12,
            $13,
            $14,
            $15,
            $16,
            $17,
            $18,
            $19,
            $20,
            $21,
            $22,
            $23,
            $24,
            $25
            
         ) on conflict (id) do nothing;
        ' 
        ) using
        new.id,
        flag,
        new.store,
        new.scrollid,
        new.guid,
        (new.event #>>'{measurements,0,uuid}')::uuid,
        new.ctime,
        now(),
        (new.event #>> '{measurements,0,waypoints,catch,time}')::timestamptz,
        (new.event #>> '{measurements,0,waypoints,lost,time}')::timestamptz,
        new.origin,
        (new.event #>> '{measurements,0,waypoints,best,location,lane}'),
        new.doctype,
        _doclass,
        new.event #>> '{evidence,args,profile,id}',
        (new.event #>> '{measurements,0,waypoints,best,vehicle,license-plates,0,text,ucode}'),
        upper(new.event #>> '{measurements,0,waypoints,best,vehicle,license-plates,0,type,country,alpha-2}'),
        (new.event #>> '{measurements,0,waypoints,best,vehicle,license-plates,0,type,country,subdivision,code}'),
        _sp,
        _dir,
        upper(new.event #>> '{measurements,0,waypoints,best,vehicle,type,make}'),
        upper(new.event #>> '{measurements,0,waypoints,best,vehicle,type,model}'),
        coalesce(upper(new.event #>> '{measurements,0,waypoints,best,vehicle,type,make}'),'')||coalesce(' '||upper(new.event #>> '{measurements,0,waypoints,best,vehicle,type,model}'),''),
        upper(new.event #>> '{measurements,0,waypoints,best,vehicle,type,info,0,class}'),

        _g
    ;
    if _doclass notnull and to_regclass(format('public.eventindex_%s',_doclass)) notnull then
        if clear then    
            execute format('delete from public.eventindex_%s where id = $1',_doclass) using new.id;
        end if;    
        _sc := ''; _sv := '';
        for r in select * from public.templates_table where template = 'index@'||_doclass||':'||new.doctype loop
            _sc := _sc || ' ,' || r.colname;
            _sv := _sv || ' ,' || '($2 #>> ''' || r.colpath || ''')' || coalesce('::' || r.coltype , '');
        end loop;
        if not _sc = '' then
            execute format (
                'insert into public.eventindex_%s ( id %s ) values ( $1 %s ) on conflict (id) do nothing', 
                _doclass, _sc, _sv 
            ) using new.id, new.event;
        end if;    
    end if;
    for _t in select unnest(_o('event_cache_templates')::text[]) loop
        _j = object_template(_t, _doclass, new.doctype, row_to_json(new)::jsonb);
        if _j notnull then
            insert into public.eventcache (id,template,event) values ( new.id, _t, _j );
        end if;
    end loop;
    insert into public.eventdist values(new.id, new.dist);
end;
$$ language plpgsql;

create or replace function public.postfactum_event_index(new public.eventstore, out _errcode integer, out _errmsg varchar) as $$
begin
    _errcode := 0; _errmsg := 'OK';
    perform public._event_reindex( new, false );
end;
$$ language plpgsql;

--
--  F:P: postfactum violation detector 
--

create or replace function public.postfactum_event_check(new public.eventstore, out _errcode integer, out _errmsg varchar) as $$
begin
    _errcode := 0; _errmsg := 'OK';

end;
$$ language plpgsql;

--
--  F:P: postfactum notifier
--

create or replace function public.postfactum_event_notify(new public.eventstore, out _errcode integer, out _errmsg varchar) as $$
begin
    _errcode := 0; _errmsg := 'OK';

end;
$$ language plpgsql;

--
--  F:P: postfactum export
--

create or replace function public.postfactum_event_export(new public.eventstore, out _errcode integer, out _errmsg varchar) as $$
begin
    _errcode := 0; _errmsg := 'OK';

end;
$$ language plpgsql;


--
--  F:P: postfactum image processor
--

create or replace function public.postfactum_image_process(new public.eventstore, out _errcode integer, out _errmsg varchar) as $$
begin
    _errcode := 0; _errmsg := 'OK';
end;
$$ language plpgsql;


