CREATE or replace FUNCTION plate_proposition(
	imoment timestamptz, 
	ipost uuid, 
	ichannel integer, 
	idirection integer, 
	iplate text, 
	ileft float8, 
	itop float8, 
	iright float8, 
	ibottom float8, 
	OUT verdict text, OUT diagnosis text) RETURNS record
    LANGUAGE plpgsql
    AS $$
declare
	filter_positive_template		text[]   := _o('platear_filter_positive_template')		 ::text[];  	   
	filter_negative_template		text[]   := _o('platear_filter_negative_template')       ::text[];     
	filter_frequency_limit			numeric  := _o('platear_filter_frequency_limit')      	 ::numeric; 	   
	filter_unistring_limit			integer  := _o('platear_filter_unistring_limit')      	 ::integer; 	   
	filter_size_deviation_limit		numeric  := _o('platear_filter_size_deviation_limit')    ::numeric;    
	filter_position_deviation_limit	numeric  := _o('platear_filter_position_deviation_limit')::numeric; 
	damper_number_timeout			interval := _o('platear_damper_number_timeout')      	 ::interval;	   
	damper_position_timeout			interval := _o('platear_damper_position_timeout')      	 ::interval;	   
	damper_number_distance			integer  := _o('platear_damper_number_distance')      	 ::integer;	   
	damper_position_distance		numeric  := _o('platear_damper_position_distance')     	 ::numeric;    
	damper_position_deviation_limit	numeric  := _o('platear_damper_position_deviation_limit')::numeric; 

	i integer;
	_m numeric;
	_s numeric;
	_x integer;
	_y integer;
	_box box;
	sect text;

begin
	verdict := 'ACCEPT'; diagnosis := 'OK';
	sect := ichannel::text||'@'||ipost::text;

	if not exists ( select id from channel where post = ipost and channel = ichannel ) then
		verdict := 'DECLINE'; diagnosis := 'Invalid source';
		return;
	end if;

	-- check allow template
	if filter_positive_template is distinct from '{}'::text[] then
		for i in array_lower(filter_positive_template, 1) .. array_upper(filter_positive_template, 1) loop
			if iplate ~ filter_positive_template[i] then
				verdict := 'ACCEPT';
				diagnosis := 'Allowed by template atom: '||filter_positive_template[i];
				return;
			end if;	
		end loop;
	end if;

	-- check decline template
	if filter_negative_template is distinct from '{}'::text[] then
		for i in array_lower(filter_negative_template, 1) .. array_upper(filter_negative_template, 1) loop
			if iplate ~ filter_negative_template[i] then
				verdict := 'DECLINE';
				diagnosis := 'Declined by template atom: '||filter_negative_template[i];
				return;
			end if;	
		end loop;
	end if;

	-- rebuild deviation stats
	if ( 
		   filter_size_deviation_limit > 0.0 
		or filter_position_deviation_limit > 0.0 
		or (
			    damper_position_timeout > '0'::interval 
			and damper_position_distance = 0.0 
			and damper_position_deviation_limit > 0.0
		)
	) then
		perform channel_restat (ipost, ichannel, 'plate.deviation');
	end if;

	-- check size deviation
	if filter_size_deviation_limit > 0 then
		_m := core.getopt(sect,'stat.size.mean')::numeric;
		_s := core.getopt(sect,'stat.size.deviation')::numeric;
		if _m notnull and _m > 0.0 and _s notnull and _s > 0.0 then
			if abs((iright - ileft)::numeric - _m) > _s * filter_size_deviation_limit then
				verdict := 'DECLINE'; diagnosis := 'Plate size deviation exceeded';
				return;
			end if;
		end if;	
	end if;


	-- check position deviation
	if filter_position_deviation_limit > 0.0 then
		_x := core.getopt(sect,'stat.position.mean.x')::integer;
		_y := core.getopt(sect,'stat.position.mean.y')::integer;
		_s := core.getopt(sect,'stat.position.deviation')::numeric;
		if _x notnull and _x > 0 and _y notnull and _y > 0 and _s notnull and _s > 0.0 then
			_box := box(point(ileft::float8, itop::float8),point(iright::float8, ibottom::float8));
			if (point(_box) <-> point(_x::float8, _y::float8))/width(_box) > filter_position_deviation_limit * _s then
				verdict := 'DECLINE'; diagnosis := 'Plate position deviation exceeded';
				return;
			end if;
		end if;	
	end if;

	-- check char frequency
	if filter_frequency_limit > 0 then
		select into _m max(freq)::numeric from lib.stringz_bar_chart(iplate);
		if _m notnull and _m > 0.0 then
			if _m > filter_frequency_limit then
				verdict := 'DECLINE'; diagnosis := 'Character relative frequency exceeded';
				return;
			end if;
		end if;	
	end if;

	-- check unistring
	if filter_unistring_limit > 0 then
		select into _m max(strlen)::numeric from lib.stringz_split_uniform(iplate);
		if _m notnull and _m > 0.0 then
			if _m > filter_unistring_limit::numeric then
				verdict := 'DECLINE'; diagnosis := 'Uniform substring length exceeded';
				return;
			end if;
		end if;	
	end if;

	-- check damper number
	if damper_number_timeout > '0'::interval then
		select into i id 
		from platear 
		where post_id = ipost 
			and channel = ichannel
			and (idirection = 0 or direction = idirection)
			and ctime < imoment 
			and ctime > imoment - damper_number_timeout 
			and lib.ilevenstein(number,iplate) <= damper_number_distance
		order by ctime desc limit 1;	
		if found and i notnull and i > 0 then
				verdict := 'DECLINE'; diagnosis := 'Damped by number duplicate from id:'||i::text;
				return;
		end if;	
	end if;

	-- check damper position
	if damper_position_timeout > '0'::interval then
		if damper_position_distance > 0.0 then
			_s :=  damper_position_distance;
		elsif damper_position_deviation_limit > 0.0 then
			_s := core.getopt(sect,'stat.position.deviation')::numeric;
			if _s notnull and _s > 0.0 then
				_s := _s * damper_position_deviation_limit;
			end if;
		end if;
		if factor notnull and factor > 0.0 then
			select into i id from platear
			where post_id = ipost 
				and channel = ichannel
				and (idirection = 0 or direction = idirection)
				and ctime < imoment
				and ctime > imoment - damper_position_timeout
				and
					(
						box(point(ileft::float8,itop::float8),point(iright::float8,ibottom::float8)) <->
						box(point(rect_left::float8,rect_top::float8),point(rect_right::float8,rect_bottom::float8))
					) / width (
						box(point(ileft::float8,itop::float8),point(iright::float8,ibottom::float8))
					) < _s 
			order by ctime desc limit 1;
			if found and i notnull and i > 0 then
					verdict := 'DECLINE'; diagnosis := 'Damped by constant distance from id:'||i::text;
					return;
			end if;	
		end if;
	end if;

end;
$$;

